==> bordeaux-0.8.1.tar.gz: bordeaux-0.8.1/src/server.cpp <==
/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <moderncli/args.hpp>
#include <moderncli/keyfile.hpp>
#include <moderncli/strings.hpp>
#include <moderncli/filesystem.hpp>

#include <atomic>
#include <thread>
#include <chrono>
#include <fstream>
#include <iostream>
#include <csignal>
#include <climits>
#include <cstdio>
#include <new>

#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <pwd.h>
#include <mqueue.h>

#include "server.hpp"
#include "utils.hpp"
#include "ipc.hpp"
#include "logger.hpp"
#include "network.hpp"
#include "drivers.hpp"
#include "sessions.hpp"

#ifdef  SYSTEMD_SERVICE
#include <systemd/sd-daemon.h>
#else
inline auto sd_notify(int, const char *) {  // NOLINT
    return 0;
}
#endif

using namespace server;

std::atomic<bool> server::running = false;  // need not be atomic...but...
std::mutex server::loglock;
unsigned server::logging = 0;
bool server::ipv6 = false;

namespace {
auto driver_name = BORDEAUX_DEFAULT_DRIVER;
const char *config_file = SERVER_CONFIG_FILE;
const char *prefix_path = SERVER_PREFIX_PATH;

args::option opt_ipv6('6',  "--ipv6", "enable ipv6");
args::option opt_config(0,  "--config", "configuration file", "file");
args::option opt_driver('d',"--driver", "driver name to load", "path");
args::option helpflag('h',  "--help", "display this list");
#ifdef  DEBUG
args::option opt_module('m',"--module", "module path to load", "path");
#endif
args::option paths(0,       "--paths", "display paths used");
args::option opt_prefix(0,  "--prefix", "prefix directory", "path");
args::option verbose('v',   "--verbose", "verbosity, repeat for debug", true);
args::option version(0,     "--version", "display version information");
args::option althelp('?',   nullptr, nullptr);

auto reload() {
    keyfile config({config_file, "custom.conf"});
    auto& hostname = config["server"]["hostname"];
    auto& username = config["server"]["username"];
    auto& media = config["server"]["media"];
    auto& language = config["server"]["language"];
    auto& location = config["server"]["location"];
    auto& voice = config["server"]["voice"];
    auto& ring_timer = config["timers"]["ring"];
    auto& answer_timer = config["timers"]["answer"];

    auto family = lower_case(config["server"]["family"]);
    if(family == "af_inet6" || family == "inet6" || family == "ipv6" || family == "6")
        ipv6 = true;

    if(hostname.empty()) {
        auto env = util::get_env("BORDEAUX_HOSTNAME", 255);
        if(!env || !*env)
            env = util::get_env("PUBLIC_HOSTNAME", 255);

        if(env && *env)
            hostname = env;
        else
            hostname = net::hostname();
    }

    if(language.empty()) {
        auto env = util::get_env("BORDEAUX_LANGUAGE", 16);
        if(!env || !*env)
            env = util::get_env("LANGUAGE", 16);

        if(!env || !*env)
            env = "en_US";

        language = env;
    }

    if(location.empty()) {
        auto env = util::get_env("BORDEAUX_LOCATION", 64);
        if(!env || !*env)
            env = util::get_env("LOCATION", 64);

        if(!env || !*env)
             env = "unknown";

        location = env;
    }

    if(voice.empty()) {
        auto env = util::get_env("BORDEAUX_VOICE", 64);
        if(!env || !*env)
            env = util::get_env("VOICE", 64);

        if(!env || !*env)
            env = "default";

        voice = env;
    }

    setenv("LANGUAGE", language.c_str(), 1);
    setenv("LOCATION", location.c_str(), 1);
    setenv("VOICE", voice.c_str(), 1);

    auto env = util::get_env("BORDEAUX_SESSIONS", 16);
    if(env && *env)
        config["server"]["limit"] = env;

    if(hostname.empty())
        hostname = net::hostname();

    if(ring_timer.empty())
        ring_timer = "4";

    if(answer_timer.empty())
        answer_timer = "30";

    if(!media.empty())
        media = net::uri_address(media);
    else
        media = net::uri_address(hostname);

    net::fqdn(hostname);

    if(username.empty()) {
        auto pwd = getpwuid(geteuid());
        if(pwd)
            username = pwd->pw_name;
        else
            username = "bordeaux";
        endpwent();
    }

    running = true;
    Plugin::config(config);
    return Driver::config_instance(config);
}

auto stop() {
    sd_notify(1, "STOPPING=1");
    running = false;
    std::this_thread::sleep_for(std::chrono::milliseconds(50));

    Driver::drop_instance();
    Plugin::stop();
    Session::stop();
    Driver::stop_instance();
    fsys::remove(SERVER_CONTROL_IPC);
}

void process_events(const ipc_event& event) {
    switch(event.type) {
    case IPC_STATUS:
        break;
    case IPC_LEVEL:
        logging = event.body.level;
        log_notice("logging level changed to {}", logging);
        break;
    case IPC_RELOAD:
        raise(SIGHUP);
        break;
    default:
        debug(4, "received unknown message {}", static_cast<int>(event.type));
    }
}

auto loader(const std::string& prefix, const std::string& name, const std::string& type) {
    auto module = type;
    module += name;
    module += ".so";
    auto dsopath = prefix + "/" + module;
    auto ptr = dlopen(dsopath.c_str(), RTLD_NOW | RTLD_GLOBAL);

#ifdef  DEBUG
    if(!ptr && name.find('/') != std::string::npos) {
        if(name[0] == '/')
            dsopath = name;
        else
            dsopath = ::getenv("PWD") + "/"s + name;    // FlawFinder: ignore
        ptr = dlopen(dsopath.c_str(), RTLD_NOW | RTLD_GLOBAL);
    }
#endif

    if(ptr)
        log_info("loading {}", dsopath);
    else
        log_error("{}: failed to load", name);
    return ptr != nullptr;
}
} // anon namespace

auto main(int argc, const char **argv) -> int { // NOLINT
#ifdef  DEBUG
    char pwd[512];
    getcwd(pwd, sizeof(pwd));
    setenv("PWD", pwd, 1);
    remove(SERVER_LOGGER_PATH "/bordeaux.log");
#endif

    auto exit_code = 0;
    log_reset();

    atexit([]{
        if(running) {
            stop();
            log_warn("exiting server");
        }
        mq_unlink(BORDEAUX_IPCPATH);
        shm_unlink(BORDEAUX_IPCPATH);
    });

    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count) {
            args::help({SERVER_ID " [options]"}, {
                SERVER_ID " " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Start Bordeaux Service"
            });
            return 0;
        }

        if(is(opt_config))
            config_file = *opt_config;

        if(is(opt_prefix))
            prefix_path = *opt_prefix;

        if(is(version)) {
            print("Version: {}\n", PROJECT_VERSION);
            return 0;
        }

        if(is(paths)) {
            print("Config:  {}\nPrefix:  {}\nDrivers: {}\nPlugins: {}\nModules: {}\nScripts: {}\nPrompts: {}\nVoices:  {}\n",
                config_file, prefix_path,
                BORDEAUX_DRIVERS_PATH, BORDEAUX_PLUGINS_PATH,
                BORDEAUX_MODULES_PATH, BORDEAUX_SCRIPTS_PATH,
                BORDEAUX_PROMPTS_PATH, BORDEAUX_VOICES_PATH);
            return 0;
        }

        // start mqueue
        struct mq_attr attr{}, old_attr{};
        memset(&attr, 0, sizeof(attr));
        attr.mq_maxmsg = 10;
        attr.mq_msgsize = sizeof(ipc_event);
        auto mq = mq_open (BORDEAUX_IPCPATH, O_RDWR | O_CREAT, 0660, &attr);
        if(mq == MQD_INVALID)
            throw std::runtime_error(format("message queue failure {}", errno));

        // set miscellaneous configs
        if(is(opt_ipv6))
            ipv6 = true;

        if(is(opt_driver))
            driver_name = *opt_driver;

        umask(007);    // FlawFinder: group access even as root...

        // create default prompt workspaces, clean old session recordings....
        fsys::current_path(prefix_path);
        fsys::remove_all("temp");
        if(fsys::create_directory("prompts"))
            fsys::permissions("prompts", fsys::perms::owner_all | fsys::perms::group_read | fsys::perms::group_write | fsys::perms::group_exec);
        if(fsys::create_directory("temp"))
            fsys::permissions("temp", fsys::perms::owner_all);

        // initial logging
        logging = verbose.count();
        log_open(SERVER_ID, LOG_INFO);

        // handlers for new and exit
        std::set_new_handler([]() {
            die(9, "error: out of memory\n");
        });

        // signal masks, we could add SIGUSR1..2 for special things
        sigset_t sigs;
        sigemptyset(&sigs);
        sigaddset(&sigs, SIGALRM);
        sigaddset(&sigs, SIGINT);
        sigaddset(&sigs, SIGHUP);
        sigaddset(&sigs, SIGKILL);
        sigaddset(&sigs, SIGCHLD);
        sigaddset(&sigs, SIGTERM);
        sigaddset(&sigs, SIGUSR1);
        pthread_sigmask(SIG_BLOCK, &sigs, nullptr);

        // bind driver
        if(!loader(BORDEAUX_DRIVERS_PATH, driver_name, "driver_"))
            die(-1, "error: driver {} failed: {}\n", driver_name, dlerror());

        // load plugins
        scan_directory(BORDEAUX_PLUGINS_PATH, [](const auto& entry) {
            if (!entry.is_regular_file())
                return false;

            auto filename = entry.path().filename().string();
            if(begins_with(filename, "plugin_") && ends_with(filename, ".so")) {
                filename.erase(filename.size() - 3, 3);
                filename.erase(0, 7);
                loader(BORDEAUX_PLUGINS_PATH, filename, "plugin_");
                return true;
            }
            return false;
        });

#ifdef  DEBUG
        if(is(opt_module))
            loader("/noexist", *opt_module, "plugin_");
#endif

        // start running service
        auto config = reload();

        Driver::start_instance(config);
        Plugin::start(config);

        // create ipc info
        fsys::remove("ipc.json");
        std::ofstream ipc("ipc.json");
        ipc << R"({"ipc_path":")" << BORDEAUX_IPCPATH << R"(",)" << std::endl
            << R"( "event_size":)" << sizeof(ipc_event) << "," << std::endl
            << R"( "system_size":)" << sizeof(ipc_system) << "," << std::endl
            << R"( "session_size":)" << sizeof(ipc_session) << "," << std::endl
            << R"( "session_count":)" << Driver::instance()->limit() << std::endl
            << "}" << std::endl;
        ipc.close();

        // create timers and other threads
        std::thread session_timers(&Session::start, 2);
        std::this_thread::sleep_for(std::chrono::milliseconds(50));

        // system notifiers
        struct sigevent mq_event{};
        mq_event.sigev_notify = SIGEV_SIGNAL;
        mq_event.sigev_signo = SIGUSR1;
        mq_notify(mq, &mq_event);
        attr.mq_flags = O_NONBLOCK;
        mq_setattr(mq, &attr, &old_attr);
        ipc_event event{};
        auto pri = 0U;

        // handle signals while up
        auto timer = sync_clock() + std::chrono::seconds(60);
        if(!eq(driver_name, "test"))
            alarm(60);
        log_info("started server");
        sd_notify(0, "READY=1");
        while(running) {
            int sig{};
            if(sigwait(&sigs, &sig))
                break;
            switch(sig) {
            case SIGCHLD:
                for(;;) {
                    auto status = 0;
                    auto pid = waitpid(-1, &status, WNOHANG);
                    if(pid == -1)
                        break;
                }
                break;
            case SIGINT:
                std::cout << '\n';
                stop();
                break;
            case SIGALRM:
                // abort if testing..
                if(eq(driver_name, "test")) {
                    exit_code = 5;
                    stop();
                }
                // sync alarm every 60 seconds, used for stats...
                sd_notify(0, "WATCHDOG=1");
                debug(3, "minute timer wakeup");
                timer += std::chrono::seconds(60);
                alarm(std::chrono::duration_cast<std::chrono::seconds>(timer - sync_clock()).count());
                break;
            case SIGTERM:
            case SIGQUIT:
                stop();
                break;
            case SIGHUP:
                sd_notify(0, "RELOADING=1");
                log_reset();
                log_info("reloading server");
                config = reload();
                Session::sync_store();
                Plugin::reload(config);
                Driver::reload_instance(config);
                sd_notify(0, "READY=1");
                break;
            case SIGUSR1:
                while(mq_receive(mq, reinterpret_cast<char *>(&event), sizeof(event), &pri) != -1)
                    process_events(event);
                mq_notify(mq, &mq_event);
                break;
            default:
                break;
            }
        }

        // normal teardown and exit
        session_timers.join();
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: {} [options]\nerror: {}\n", SERVER_ID, e.what());
    }
    catch(const std::exception& e) {
        if(running)
            log_fail(-1, "{}", e.what());
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(running)
            log_fail(-1, "unknown failure");
        else
            die(-1, "error: unknown failure\n");
    }

    log_info("stopped server; reason={}", exit_code);
    return exit_code;
}

==> bordeaux-v0.9.2.tar.gz: bordeaux-v0.9.2/src/server.cpp <==
/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"       // IWYU pragma: keep
#include "utils.hpp"
#include "ipc.hpp"
#include "network.hpp"
#include "drivers.hpp"
#include "sessions.hpp"

#include <moderncli/args.hpp>
#include <moderncli/keyfile.hpp>
#include <moderncli/filesystem.hpp>
#include <moderncli/process.hpp>
#include <moderncli/socket.hpp>

#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <mqueue.h>

#ifdef  SYSTEMD_SERVICE
#include <systemd/sd-daemon.h>
#else
inline auto sd_notify(int, const char *) {  // NOLINT
    return 0;
}
#endif

using namespace server;

std::atomic<bool> server::running = false;  // need not be atomic...but...
system_logger server::logger;
bool server::ipv6 = false;

namespace {
auto driver_name = BORDEAUX_DEFAULT_DRIVER;
fsys::path config_file = SERVER_CONFIG_FILE;
fsys::path prefix_path = SERVER_PREFIX_PATH;

args::option opt_ipv6('6',  "--ipv6", "enable ipv6");
args::option opt_config(0,  "--config", "configuration file", "file");
args::option opt_driver('d',"--driver", "driver name to load", "path");
args::option helpflag('h',  "--help", "display this list");
args::option paths(0,       "--paths", "display paths used");
args::option opt_prefix(0,  "--prefix", "prefix directory", "path");
args::option verbose('v',   "--verbose", "verbosity, repeat for debug", true);
args::option version(0,     "--version", "display version information");
args::option althelp('?',   nullptr, nullptr);
unsigned logging{};

auto is_root() {
    return getuid() == 0 || getgid() == 0;
}

auto reload() {
    keyfile config;
    config["consts"];
    config["paths"]["scripts"] = BORDEAUX_SCRIPTS_PATH;
    config["paths"]["voices"] = BORDEAUX_VOICES_PATH;
    config["paths"]["prompts"] = BORDEAUX_PROMPTS_PATH;
    config["paths"]["modules"] = BORDEAUX_MODULES_PATH;
    config["paths"]["prefix"] = SERVER_PREFIX_PATH;
    config["paths"]["logger"] = SERVER_LOGGER_PATH;

    config.load(config_file);
    config.load("custom.conf");

    auto& hostname = config["server"]["hostname"];
    auto& media = config["server"]["media"];
    auto& language = config["server"]["language"];
    auto& location = config["server"]["location"];
    auto& voice = config["server"]["voice"];
    auto& ring_timer = config["timers"]["ring"];
    auto& answer_timer = config["timers"]["answer"];

    auto family = lower_case(config["server"]["family"]);
    if(family == "af_inet6" || family == "inet6" || family == "ipv6" || family == "6")
        ipv6 = true;

    if(hostname.empty())
        hostname = system_hostname();

    if(language.empty()) {
        auto env = process::env("BORDEAUX_LANGUAGE", 16);
        if(!env || (*env).empty())
            env = process::env("LANGUAGE", 16);

        if(!env || (*env).empty())
            language = "en_US";
        else
            language = *env;
    }

    if(location.empty()) {
        auto env = process::env("BORDEAUX_LOCATION", 64);
        if(!env || (*env).empty())
            env = process::env("LOCATION", 64);

        if(env)
            location = *env;
        else
            location = "unknown";
    }

    if(voice.empty()) {
        auto env = process::env("BORDEAUX_VOICE", 64);
        if(!env || (*env).empty())
            env = process::env("VOICE", 64);

        if(!env || (*env).empty())
            env = "default";
        else
            voice = *env;
    }

    process::env("LANGUAGE", language);
    process::env("LOCATION", location);
    process::env("VOICE", voice);

    if(hostname.empty())
        hostname = system_hostname();

    if(ring_timer.empty())
        ring_timer = "4";

    if(answer_timer.empty())
        answer_timer = "30";

    if(!media.empty())
        media = net::uri_address(media);
    else
        media = net::uri_address(hostname);

    // Support for testability... will be extended over time
    if(!is_root() && is(opt_prefix) && fsys::exists(prefix_path / "bordeaux.conf"s)) {
        config["paths"]["scripts"] = prefix_path;
        config["paths"]["modules"] = prefix_path;
        config["paths"]["voices"] = fsys::canonical(prefix_path / ".." / "voices");
        config["paths"]["prompts"] = prefix_path / "prompts";
        config["paths"]["prefix"] = prefix_path;
        config["paths"]["logger"] = prefix_path;
    }

    hostname = inet_host(hostname);
    Plugin::config(config);
    return Driver::config_instance(config, logging);
}

auto stop() {
    sd_notify(1, "STOPPING=1");
    running = false;
    std::this_thread::sleep_for(std::chrono::milliseconds(50));

    Driver::drop_instance();
    Plugin::stop();
    Session::stop();
    Driver::stop_instance();
    fsys::remove(SERVER_CONTROL_IPC);
}

void process_events(const ipc_event& event) {
    switch(event.type) {
    case IPC_STATUS:
        break;
    case IPC_LEVEL:
        logging = event.body.level;
        logger.notice("logging level changed to {}", logging);
        break;
    case IPC_RELOAD:
        raise(SIGHUP);
        break;
    default:
        logger.debug(4, "received unknown message {}", static_cast<int>(event.type));
    }
}

auto loader(const std::string& prefix, const std::string& name, const std::string& type) {
    auto module(type);
    module += name;
    module += process::dso_suffix;
    auto dsopath = prefix + "/" + module;
    if(name[0] == '/' && !is_root())
        dsopath = name + process::dso_suffix;
    auto ptr = process::load(dsopath);

    if(ptr)
        logger.info("loading {}", dsopath);
    else
        logger.error("{}: failed to load", name);
    return ptr != nullptr;
}

void cleanup() {
    running = false;
    mq_unlink(BORDEAUX_IPCPATH);
    shm_unlink(BORDEAUX_IPCPATH);
}
} // anon namespace

auto main(int argc, const char **argv) -> int { // NOLINT
    auto exit_code = 0;
    log_reset();
    on_crit(cleanup);

    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count) {
            args::help({SERVER_ID " [options]"}, {
                SERVER_ID " " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Start Bordeaux Service"
            });
            return 0;
        }

        if(is(opt_config))
            config_file = *opt_config;

        if(is(opt_prefix))
            prefix_path = fsys::canonical(*opt_prefix);

        if(is(version)) {
            print("Version: {}\n", PROJECT_VERSION);
            return 0;
        }

        if(is(paths)) {
            print("Config:  {}\nPrefix:  {}\nDrivers: {}\nPlugins: {}\nModules: {}\nScripts: {}\nPrompts: {}\nVoices:  {}\n",
                config_file, prefix_path,
                BORDEAUX_DRIVERS_PATH, BORDEAUX_PLUGINS_PATH,
                BORDEAUX_MODULES_PATH, BORDEAUX_SCRIPTS_PATH,
                BORDEAUX_PROMPTS_PATH, BORDEAUX_VOICES_PATH);
            return 0;
        }

        // start mqueue
        struct mq_attr attr{}, old_attr{};
        memset(&attr, 0, sizeof(attr));
        attr.mq_maxmsg = 10;
        attr.mq_msgsize = sizeof(ipc_event);
        auto mq = mq_open (BORDEAUX_IPCPATH, O_RDWR | O_CREAT, 0660, &attr);
        if(mq == MQD_INVALID)
            throw std::runtime_error(format("message queue failure {}", errno));

        // set miscellaneous configs
        if(is(opt_ipv6))
            ipv6 = true;

        if(is(opt_driver))
            driver_name = *opt_driver;

        umask(007);    // FlawFinder: group access even as root...

        // create default prompt workspaces, clean old session recordings....
        fsys::current_path(prefix_path);
        process::env("PWD", prefix_path);
        fsys::remove_all("temp");
        if(fsys::create_directory("prompts"))
            fsys::permissions("prompts", fsys::perms::owner_all | fsys::perms::group_read | fsys::perms::group_write | fsys::perms::group_exec);
        if(fsys::create_directory("temp"))
            fsys::permissions("temp", fsys::perms::owner_all);

        // initial logging
        logging = verbose.count();
        logger.set(verbose.count(), logger_notify);
        system_logger::open(SERVER_ID, LOG_INFO);   // FlawFinder: ignore

        // handlers for new and exit
        std::set_new_handler([]() {
            crit(9, "error: out of memory\n");
        });

        // signal masks, we could add SIGUSR1..2 for special things
        sigset_t sigs;
        sigemptyset(&sigs);
        sigaddset(&sigs, SIGALRM);
        sigaddset(&sigs, SIGINT);
        sigaddset(&sigs, SIGHUP);
        sigaddset(&sigs, SIGKILL);
        sigaddset(&sigs, SIGCHLD);
        sigaddset(&sigs, SIGTERM);
        sigaddset(&sigs, SIGUSR1);
        pthread_sigmask(SIG_BLOCK, &sigs, nullptr);

        // bind driver
        auto driver_path = fsys::path(BORDEAUX_DRIVERS_PATH);
        auto dp = driver_name;
        if(!is_root() && strchr(driver_name, '/') && driver_name[0] != '/') {
            driver_path = fsys::canonical(prefix_path / ".." / fsys::path(driver_name).parent_path());
            dp = strrchr(driver_name, '/');
            if(dp) {
                ++dp;
                auto cp = strchr(driver_name, '_');
                if(cp)
                    dp = ++cp;
            }
            else
                dp = driver_name;
        }
        if(!loader(driver_path, dp, "driver_"))
            crit(-1, "error: driver {} failed: {}\n", driver_name, dlerror());

        // load plugins if not absolute driver for build testing
        auto plugins_path = fsys::path(BORDEAUX_PLUGINS_PATH);
        if(!is_root() && strchr(driver_name, '/')) {
            if(driver_name[0] == '/')
                plugins_path = fsys::path(driver_name).parent_path();
            else
                plugins_path = prefix_path / ".." / fsys::path(driver_name).parent_path();
            if(fsys::is_directory(plugins_path / "../plugins"))
                plugins_path = fsys::canonical(plugins_path / "../plugins");
            else
                plugins_path = fsys::canonical(plugins_path / "../../plugins" / plugins_path.filename());
        }

        if(fsys::is_directory(plugins_path))
            scan_directory(plugins_path, [plugins_path](const auto& entry) {
                if (!entry.is_regular_file())
                    return false;

                auto filename = entry.path().filename().string();
                if(begins_with(filename, "plugin_") && ends_with(filename, ".so")) {
                    filename.erase(filename.size() - 3, 3);
                    filename.erase(0, 7);
                    loader(plugins_path, filename, "plugin_");
                    return true;
                }
                return false;
            });

        // start running service
        auto config = reload();
        running = true;

        Driver::start_instance(config);
        Plugin::start(config);

        // create ipc info
        fsys::remove("ipc.json");
        std::ofstream ipc("ipc.json");
        ipc << R"({"ipc_path":")" << BORDEAUX_IPCPATH << R"(",)" << std::endl
            << R"( "event_size":)" << sizeof(ipc_event) << "," << std::endl
            << R"( "system_size":)" << sizeof(ipc_system) << "," << std::endl
            << R"( "session_size":)" << sizeof(ipc_session) << "," << std::endl
            << R"( "session_count":)" << Driver::instance()->limit() << std::endl
            << "}" << std::endl;
        ipc.close();

        // create timers and other threads
        std::thread session_timers(&Session::start, 2);
        std::this_thread::sleep_for(std::chrono::milliseconds(50));

        // system notifiers
        struct sigevent mq_event{};
        mq_event.sigev_notify = SIGEV_SIGNAL;
        mq_event.sigev_signo = SIGUSR1;
        mq_notify(mq, &mq_event);
        attr.mq_flags = O_NONBLOCK;
        mq_setattr(mq, &attr, &old_attr);
        ipc_event event{};
        auto pri = 0U;

        // handle signals while up
        auto timer = sync_clock() + std::chrono::seconds(60);
        if(!eq(driver_name, "test"))
            alarm(60);
        logger.info("started server");
        sd_notify(0, "READY=1");
        while(running) {
            int sig{};
            if(sigwait(&sigs, &sig))
                crit(-1, "Broken sigwait");

            switch(sig) {
            case SIGCHLD:
                for(;;) {
                    auto status = 0;
                    auto pid = waitpid(-1, &status, WNOHANG);
                    if(pid == -1)
                        break;
                }
                break;
            case SIGINT:
                std::cout << '\n';
                stop();
                break;
            case SIGALRM:
                // abort if testing..
                if(eq(driver_name, "test")) {
                    exit_code = 5;
                    stop();
                }
                // sync alarm every 60 seconds, used for stats...
                sd_notify(0, "WATCHDOG=1");
                logger.debug(3, "minute timer wakeup");
                timer += std::chrono::seconds(60);
                alarm(std::chrono::duration_cast<std::chrono::seconds>(timer - sync_clock()).count());
                break;
            case SIGTERM:
            case SIGQUIT:
                stop();
                break;
            case SIGHUP:
                sd_notify(0, "RELOADING=1");
                log_reset();
                logger.info("reloading server");
                Session::sync_store();
                try {
                    config = reload();
                    Plugin::reload(config);
                    Driver::reload_instance(config);
                }
                catch(const std::exception& e) {
                    logger.error("reload: ", e.what());
                }
                catch(...) {
                    logger.error("reload: unknown error");
                }
                sd_notify(0, "READY=1");
                break;
            case SIGUSR1:
                while(mq_receive(mq, reinterpret_cast<char *>(&event), sizeof(event), &pri) != -1)
                    process_events(event);
                mq_notify(mq, &mq_event);
                break;
            default:
                break;
            }
        }

        // normal tear-down and exit
        session_timers.join();
    }
    catch(const bad_arg& e) {
        crit(-1, "Usage: {} [options]\nerror: {}\n", SERVER_ID, e.what());
    }
    catch(const std::exception& e) {
        if(running)
            logger.crit(-1, "{}", e.what());
        else
            crit(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(running)
            logger.crit(-1, "unknown failure");
        else
            crit(-1, "error: unknown failure\n");
    }

    logger.info("stopped server; reason={}", exit_code);
    cleanup();
    return exit_code;
}
